#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <stdarg.h>

#include "common.h"

#ifndef WIN32
    #define _getch getch
    #define _kbhit kbhit
#endif

#ifdef BC30
	#include <dos.h>
#elif defined(DJGPP) || defined(LINUX)
    //#include <bios.h>
    #include <time.h>
#endif


/*****************  Common Message Functions  ******************/
int   ERR_Type;
char* ERR_Msg;

static char  ERR_MsgBuf[512];
static FILE* MSG_outputDevice = stdout;
static FILE* ERR_outputDevice = stderr;

bool msg_SetLogFile(char* logfile) {
    msg_CloseLogFile();

	if ((MSG_outputDevice = fopen(logfile, "w")) != NULL) return true;

	MSG_outputDevice = stdout;
    return false;
}

void msg_CloseLogFile(void) {
	if (ERR_outputDevice == MSG_outputDevice) ERR_outputDevice = stderr;
	
	if (MSG_outputDevice != stdout) fclose(MSG_outputDevice);
	MSG_outputDevice = stdout;
}

bool err_SetLogFile(char* logfile) {
    err_CloseLogFile();

	if (logfile == NULL)
	{
		ERR_outputDevice = MSG_outputDevice;
		return true;
	}

	if ((ERR_outputDevice = fopen(logfile, "w")) != NULL) return true;

	ERR_outputDevice = stderr;
    return false;
}

void err_CloseLogFile(void) {
    if ((ERR_outputDevice != stderr) && (ERR_outputDevice != MSG_outputDevice))
        fclose(ERR_outputDevice);

	ERR_outputDevice = stderr;
}

#ifdef _MANAGED
	#pragma managed(push, off)
#endif
void showmsg(char* fmt, ...) {
     va_list args;

	 va_start(args, fmt);
	 vfprintf(MSG_outputDevice, fmt, args);
	 va_end(args);

     fflush(MSG_outputDevice);
}

void errmsg(char* fmt, ...) {
     va_list args;

	 va_start(args, fmt);
	 vfprintf(ERR_outputDevice, fmt, args);
	 va_end(args);

     fflush(ERR_outputDevice);
}

void errmsg_exit(char* fmt, ...) {
     va_list args;

	 va_start(args, fmt);
	 vfprintf(ERR_outputDevice, fmt, args);
	 va_end(args);

     fflush(ERR_outputDevice);
	 exit(2);
}

void err_SetMsg(int errtype, char* fmt, ...) {
     va_list args;

	 va_start(args, fmt);
	 vsprintf(ERR_MsgBuf, fmt, args);
	 va_end(args);

	 ERR_Type = errtype;
	 ERR_Msg  = ERR_MsgBuf;
}
#ifdef _MANAGED
	#pragma managed(pop)
#endif
/*---------------- end of Message Functions -------------------*/



/******************  Common Timer Functions  *******************/
unsigned long timer_nowtime(void) { //in ms
#ifdef WIN32
	return GetTickCount();
#elif defined BC30
    static int usetimer = 0;
    static unsigned long far* timeraddr;
    static unsigned long inittime;
    
    if (usetimer == 0)
    {
        timeraddr = (unsigned long far*)MK_FP(0x40,0x6c);
        inittime  = *timeraddr;
        usetimer = 1;
    }
    
    return ((*timeraddr) - inittime) * 55UL;
#elif defined(DJGPP) || defined(LINUX)
    static int usetimer = 0;
    static unsigned long long inittime;
    
    if (usetimer == 0)
    {
        //inittime  = biostime(0, 0);
        inittime = uclock();
        usetimer = 1;
    }
    
    //return (biostime(0, 0) - inittime) * 55UL;
    return (unsigned long)((uclock() - inittime)*1000LL/UCLOCKS_PER_SEC);
#else
	errmsg_exit("ERROR: timer is not supported!\n");
#endif
}


void delay_ms(unsigned long delaytime) { //delay in ms
	delaytime = delaytime + timer_nowtime();
	while (timer_nowtime() < delaytime);
}
/*------------------ end of Timer Functions -------------------*/



/*****************  Common Keyboard Functions  *****************/
unsigned keyboard_getkey(void) {
    unsigned c;
    if (!_kbhit()) return 0;

    c = (unsigned)_getch();

    if ((c != 0) && (c != 0xe0))
        return (c & 0x00ff);
    else
    {
        c = (unsigned)_getch();
        return (c | 0xff00);
    }
}

unsigned keyboard_waitkey(void) {
    unsigned c;

	while ((c = keyboard_getkey()) == 0);
	return c;
}
/*----------------- end of Keyboard Functions -----------------*/



/******************  Common Memory Functions  ******************/
void* mem_alloc(size_t size) {
	void* p;

	if ((p = malloc(size)) == NULL)
	   errmsg_exit("ERROR: no enough memory to allocate %u bytes!", size);

	return p;
}

void* mem_realloc(void* pointer, size_t size) {
	void* p;

	if ((p = realloc(pointer, size)) == NULL)
	   errmsg_exit("ERROR: no enough memory to reallocate %u bytes!", size);

	return p;
}
/*------------------- end of Memory Functions -----------------*/

