#include "common.h"
#include "io.h"

#ifdef WIN32
    #define USE_WINIO
    //#define USE_PCIDEBUG
#endif


#if defined     USE_WINIO
    #include <winio.h>
#elif defined   USE_PCIDEBUG
	#include <windows.h>
    #include <pcifunc.h>
#elif defined   LINUX
    #include <sys/io.h>
#elif defined   DJGPP
    #include <pc.h>
#elif defined   BC30
    #include <conio.h>

    //386 instructions
    #define EMIT_DWORD(x) \
        __emit__((unsigned char)x);__emit__((unsigned char)(x>>8));\
        __emit__((unsigned char)(x>>16));__emit__((unsigned char)(x>>24))

    #define MOV_EAX(x) __emit__(0x66, 0xb8); EMIT_DWORD(x)

    #define MOV_VAR_EAX(x) __emit__(0x66); _asm mov WORD PTR x, ax

    #define MOV_EAX_VAR(x) __emit__(0x66); _asm mov ax, WORD PTR x

    #define OUT_EDX_EAX __emit__(0x66); _asm out dx, ax
    #define IN_EAX_EDX __emit__(0x66); _asm in ax, dx
#endif


bool IO_inUse = false;
bool io_init(void) {
    if (IO_inUse == true) return true;

	#if defined		USE_WINIO
        if (!InitializeWinIo())
        {
            err_SetMsg(ERROR_IOINITFAIL, "I/O library fails to initialize");
            return false;
        }
    #elif defined	USE_PCIDEBUG
        if (getdllstatus())
        {
            err_SetMsg(ERROR_IOINITFAIL, "I/O library fails to initialize");
            return false;
        }
    #elif defined	LINUX        
        if (iopl(3) != 0)
		{
            err_SetMsg(ERROR_IOINITFAIL, "I/O library fails to initialize");
            return false;
        }
    #endif

	IO_inUse = true;
	return IO_inUse;
}

void io_close(void) {
    if (IO_inUse == false) return;

	#ifdef USE_WINIO
        ShutdownWinIo();
    #endif

	IO_inUse = false;
}


//outport: double word
void io_outpdw(unsigned addr, unsigned long val) {
    #if   defined   USE_WINIO
	    SetPortVal(addr, val, 4);
    #elif defined	USE_PCIDEBUG
		_IoWriteLong((ULONG)addr, val);
    #elif defined	LINUX
        outl(val, addr);
    #elif defined   DJGPP
        outportl(addr, val);
    #elif defined   BC30
        _asm mov dx, WORD PTR addr
        MOV_EAX_VAR(val);
        OUT_EDX_EAX;
    #endif
}

//inport: double word
unsigned long io_inpdw(unsigned addr) {
    #if   defined   USE_WINIO
        unsigned long val;

	    GetPortVal(addr, &val, 4);
	    return (val);
    #elif defined	USE_PCIDEBUG
		return _IoReadLong((ULONG)addr);
    #elif defined	LINUX
        return inl(addr);
	#elif defined   DJGPP
        return inportl(addr);
    #elif defined   BC30
        unsigned long retval;

        _asm mov dx, WORD PTR addr
        IN_EAX_EDX;
        MOV_VAR_EAX(retval);

        return retval;
    #endif
}


//outport: word
void io_outpw(unsigned addr, unsigned val) {
    #if   defined   USE_WINIO
	    SetPortVal(addr, val, 2);
    #elif defined	USE_PCIDEBUG
		_IoWriteShort((ULONG)addr, val);
    #elif defined	LINUX
        outw(val, addr);
    #elif defined   DJGPP
        outportw(addr, val);
    #elif defined   BC30
        outpw(addr, val);
    #endif
}

//inport: word
unsigned io_inpw(unsigned addr) {
    #if   defined   USE_WINIO
        unsigned long val;

	    GetPortVal(addr, &val, 2);
	    return ((unsigned)val);
    #elif defined	USE_PCIDEBUG
		return _IoReadShort((ULONG)addr);
    #elif defined	LINUX
        return inw(addr);
    #elif defined   DJGPP
        return inportw(addr);
    #elif defined   BC30
        return inpw(addr);
    #endif
}


//outport: byte
void io_outpb(unsigned addr, unsigned char val) {
    #if   defined   USE_WINIO
	    SetPortVal(addr, val, 1);
    #elif defined	USE_PCIDEBUG
		_IoWriteChar((ULONG)addr, val);
    #elif defined	LINUX
        outb(val, addr);
    #elif defined   DJGPP
        outportb(addr, val);
    #elif defined   BC30
        outp(addr, val);
    #endif
}

//inport: byte
unsigned char io_inpb(unsigned addr) {
    #if   defined   USE_WINIO
        unsigned long val;

	    GetPortVal(addr, &val, 1);
	    return ((unsigned char)val);
    #elif defined	USE_PCIDEBUG
		return _IoReadChar((ULONG)addr);
    #elif defined	LINUX
        return inb(addr);
    #elif defined   DJGPP
        return inportb(addr);
    #elif defined   BC30
        return inp(addr);
    #endif
}



// read south bridge register
unsigned char read_sb_regb(unsigned char idx) {
	io_outpdw(0x0cf8, (0x80003800L+(unsigned long)idx) & 0xfffffffcL);
    return 0xff & (unsigned char)(io_inpdw(0x0cfc) >> ((idx & 0x03) * 8));
}

unsigned read_sb_regw(unsigned char idx) {
	io_outpdw(0x0cf8, (0x80003800L+(unsigned long)idx) & 0xfffffffcL);
    return 0xffff & (unsigned)(io_inpdw(0x0cfc) >> ((idx & 0x03) * 8));
}

unsigned long read_sb_reg(unsigned char idx) {
	io_outpdw(0x0cf8, (0x80003800L+(unsigned long)idx) & 0xfffffffcL);
    return io_inpdw(0x0cfc);
}

// write south bridge register
void write_sb_regb(unsigned char idx, unsigned char val) {
	int i = (idx & 0x03) * 8;

	io_outpdw(0x0cf8, (0x80003800L+(unsigned long)idx) & 0xfffffffcL);
	io_outpdw(0x0cfc, (io_inpdw(0x0cfc) & (~(0x000000ffL << i))) | ((unsigned long)val << i));
}

void write_sb_regw(unsigned char idx, unsigned val) {
	int i = (idx & 0x03) * 8;

	io_outpdw(0x0cf8, (0x80003800L+(unsigned long)idx) & 0xfffffffcL);
	io_outpdw(0x0cfc, (io_inpdw(0x0cfc) & (~(0x0000ffffL << i))) | ((unsigned long)val << i));
}

void write_sb_reg(unsigned char idx, unsigned long val) {
	io_outpdw(0x0cf8, (0x80003800L+(unsigned long)idx) & 0xfffffffcL);
	io_outpdw(0x0cfc, val);
}


// read north bridge register
unsigned char read_nb_regb(unsigned char idx) {
	io_outpdw(0x0cf8, (0x80000000L+(unsigned long)idx) & 0xfffffffcL);
    return 0xff & (unsigned char)(io_inpdw(0x0cfc) >> ((idx & 0x03) * 8));
}

unsigned read_nb_regw(unsigned char idx) {
	io_outpdw(0x0cf8, (0x80000000L+(unsigned long)idx) & 0xfffffffcL);
    return 0xffff & (unsigned)(io_inpdw(0x0cfc) >> ((idx & 0x03) * 8));
}

unsigned long read_nb_reg(unsigned char idx) {
	io_outpdw(0x0cf8, (0x80000000L+(unsigned long)idx) & 0xfffffffcL);
    return io_inpdw(0x0cfc);
}

// write north bridge register
void write_nb_regb(unsigned char idx, unsigned char val) {
	int i = (idx & 0x03) * 8;

	io_outpdw(0x0cf8, (0x80000000L+(unsigned long)idx) & 0xfffffffcL);
	io_outpdw(0x0cfc, (io_inpdw(0x0cfc) & (~(0x000000ffL << i))) | ((unsigned long)val << i));
}

void write_nb_regw(unsigned char idx, unsigned val) {
	int i = (idx & 0x03) * 8;

	io_outpdw(0x0cf8, (0x80000000L+(unsigned long)idx) & 0xfffffffcL);
	io_outpdw(0x0cfc, (io_inpdw(0x0cfc) & (~(0x0000ffffL << i))) | ((unsigned long)val << i));
}

void write_nb_reg(unsigned char idx, unsigned long val) {
	io_outpdw(0x0cf8, (0x80000000L+(unsigned long)idx) & 0xfffffffcL);
	io_outpdw(0x0cfc, val);
}

