#include "common.h"
#include "io.h"
#include "pwm.h"

static int pwm_INITFLAG = 0;

bool pwm_Initialize(unsigned baseaddr, int clkmode, int irqmode) {
	if (pwm_INITFLAG != 0)
	{
        err_SetMsg(ERROR_PWM_INUSE, "PWM lib was already opened");
		return false;
	}

	if (IO_inUse == false)
	{
        err_SetMsg(ERROR_PWM_NOIO, "I/O lib must be opened before PWM lib is used");
		return false;
	}

	//NOTE: base address should be selected carefully to avoid conflicts!
	if (baseaddr == 0xffff)
        pwm_SetDefaultBaseAddress();
    else
        pwm_SetBaseAddress(baseaddr);

	pwm_SetBaseClock(clkmode);
	pwm_SetIRQ(irqmode);

	pwm_DisableMultiPWM(0xffffffffL);
	pwm_DisableMultiINT(0xffffffffL);
	pwm_ClearMultiFLAG(0xffffffffL);
	pwm_UnlockMulti(0xffffffffL);

	pwm_INITFLAG = 1;
	return true;
}

void pwm_Close(void) {
	if (pwm_INITFLAG == 0) return;

	pwm_DisableMultiPWM(0xffffffffL);
	pwm_ClearMultiFLAG(0xffffffffL);
	pwm_INITFLAG = 0;
}


static unsigned long us = 10L;
void pwm_SetBaseClock(int clock) {
    if (clock == PWMCLOCK_10MHZ)
        us = 10L;
    else
        us = 50L;

    pwm_SelectClock(clock);
}


void pwm_SetPulse(int channel, unsigned long period, unsigned long duty) {
    unsigned long low_pulse_width;

    if ((channel<0) || (channel>=32) || (duty==0L) || (period<=duty))
    {
        errmsg("ERROR: invalid pulse setting for channel %d!\n", channel);
        return;
    }

    low_pulse_width = period - duty;
    pwm_SetHREG(channel, duty*us);
    pwm_SetLREG(channel, low_pulse_width*us);
}


void pwm_SetPulse_10MHZ(int channel, unsigned long period, unsigned long duty)
{
    unsigned long low_pulse_width;

    if ((channel<0) || (channel>=32) || (duty==0L) || (period<=duty))
    {
        errmsg("ERROR: invalid pulse setting!");
        return;
    }

    low_pulse_width = period - duty;
    //pwm_SetHREG(channel, duty*us/10L);
    //pwm_SetLREG(channel, low_pulse_width*us/10L);
	if (us == 10L) //verbose code for speed
	{
		pwm_SetHREG(channel, duty);
		pwm_SetLREG(channel, low_pulse_width*us);
	}
	else
	{
		pwm_SetHREG(channel, duty*5L);
		pwm_SetLREG(channel, low_pulse_width*5L);
	}
}


/*
typedef struct {
    unsigned long hreg;
    unsigned long lreg;
    unsigned long creg;
} PWMSETTING;
static PWMSETTING pwm_settings[32];
static unsigned long syncreg;
static unsigned long intreg;

void pwm_BackupAllSettings(void) {
    int i;

    syncreg = pwm_ReadSyncREG();
    pwm_LockMulti(0xffffffffL);

    for (i=0; i<32; i++)
    {
        pwm_settings[i].hreg = pwm_ReadHREG(i);
        pwm_settings[i].lreg = pwm_ReadLREG(i);
        pwm_settings[i].creg = pwm_ReadCtrlREG(i);
    }
    
    intreg  = pwm_ReadINTREG();
}

void pwm_RestoreAllSettings(void) {
    int i;

    pwm_LockMulti(0xffffffffL);
    for (i=0; i<32; i++)
    {
        pwm_SetHREG(i, pwm_settings[i].hreg);
        pwm_SetLREG(i, pwm_settings[i].lreg);
        pwm_SetCtrlREG(i, pwm_settings[i].creg);
    }

    pwm_SetINTREG(intreg);
    pwm_SetSyncREG(syncreg);
}
*/

