#ifndef __RCSERVO_H
#define __RCSERVO_H

#include "defines.h"


typedef struct
{
	unsigned long periodWidth;			//in us
	unsigned long minDutyWidth;			//in us
	unsigned long maxDutyWidth;			//in us
	unsigned long invalidDutyWidth;		//in us
	unsigned long minLowPhaseWidth;		//in us

	unsigned long cmdDuty[8];			//in us; PWM pulse commands to read feedback or change servo's parameters
	int           cmdInPlayMode;		//use by rcservo_EnterPlayMode() to read servo's feedback

	unsigned long captureResolution;	//in 0.1us; resolution for reading servo's feedback
	unsigned long captureInitDelay;     //in us; time for waiting servo to switch its signal line from IN to OUT when reading feedback
	unsigned long captureFinishDelay;   //in us; time for waiting servo to switch its signal line from OUT to IN when reading feedback
	unsigned long captureMaxWidth;      //in captureResolution; must > maximum width of servo's feedback pulse
	unsigned long ROBOARD_widthOffest;  //in captureResolution; used to fix the captured width of the feedback pulse
										//determined according to H/W design; without pull up: 124L

	int feedbackReadMode;				//method used to read servo's feedback
	int feedbackFilterWidth;			//width of the filter used by rcservo_ReadPositionDN()
	int feedbackMaxNumFail;				//allowed maximum number of feedback-reading fails

} RCSERVO_PARAMETER_t;
//-- values for RCSERVO_PARAMETER_t.feedbackReadMode
     #define RCSERVO_READPOS_NORMALMODE (0)  //don't want to use enum:p
     #define RCSERVO_READPOS_FASTMODE   (1)

extern RCSERVO_PARAMETER_t RCSERVO_servoParameters[32];


#ifdef __cplusplus
extern "C" {
#endif

void rcservo_SetServo(int channel, int servono);
//-- values for the "servono" argument of the above functions
     #define RCSERVO_DEFAULTSERVO			(0)	 //don't want to use enum:p
     #define RCSERVO_KONDO_GENERIC	        (100) 
     #define RCSERVO_KONDO_KRS786	        (101) 
     #define RCSERVO_KONDO_KRS788			(102)
     #define RCSERVO_KONDO_KRS78X			(103)
     #define RCSERVO_HITEC_GENERIC	        (200) 
     #define RCSERVO_HITEC_HSR8498			(201)


bool rcservo_Initialize(unsigned long usedchannels);
//-- values for the "usedchannels" argument of the above functions (note
//   that different values can be added to use multiple channels)
     #define RCSERVO_USECHANNEL0			(1)
     #define RCSERVO_USECHANNEL1			(1L<<1)
     #define RCSERVO_USECHANNEL2			(1L<<2)
     #define RCSERVO_USECHANNEL3			(1L<<3)
     #define RCSERVO_USECHANNEL4			(1L<<4)
     #define RCSERVO_USECHANNEL5			(1L<<5)
     #define RCSERVO_USECHANNEL6			(1L<<6)
     #define RCSERVO_USECHANNEL7			(1L<<7)
     #define RCSERVO_USECHANNEL8			(1L<<8)
     #define RCSERVO_USECHANNEL9			(1L<<9)
     #define RCSERVO_USECHANNEL10			(1L<<10)
     #define RCSERVO_USECHANNEL11			(1L<<11)
     #define RCSERVO_USECHANNEL12			(1L<<12)
     #define RCSERVO_USECHANNEL13			(1L<<13)
     #define RCSERVO_USECHANNEL14			(1L<<14)
     #define RCSERVO_USECHANNEL15			(1L<<15)
     #define RCSERVO_USECHANNEL16			(1L<<16)
     #define RCSERVO_USECHANNEL17			(1L<<17)
     #define RCSERVO_USECHANNEL18			(1L<<18)
     #define RCSERVO_USECHANNEL19			(1L<<19)
     #define RCSERVO_USECHANNEL20			(1L<<20)
     #define RCSERVO_USECHANNEL21			(1L<<21)
     #define RCSERVO_USECHANNEL22			(1L<<22)
     #define RCSERVO_USECHANNEL23			(1L<<23)
/*
     #define RCSERVO_USECHANNEL24			(1L<<24)
     #define RCSERVO_USECHANNEL25			(1L<<25)
     #define RCSERVO_USECHANNEL26			(1L<<26)
     #define RCSERVO_USECHANNEL27			(1L<<27)
     #define RCSERVO_USECHANNEL28			(1L<<28)
     #define RCSERVO_USECHANNEL29			(1L<<29)
     #define RCSERVO_USECHANNEL30			(1L<<30)
     #define RCSERVO_USECHANNEL31			(1L<<31)
*/
//-- if the above function return false, ERR_Type may be settled as the following number:
//   #define ERROR_PWM_INUSE	     (ERR_NOERROR + 300)
//   #define ERROR_PWM_NOIO		     (ERR_NOERROR + 301)
     #define ERROR_RCSERVO_INUSE	 (ERR_NOERROR + 400)

void rcservo_Close(void);


//Functions for Capture Mode:
void          rcservo_EnterCaptureMode(void);

unsigned long rcservo_ReadPosition(int channel, int cmd);
void          rcservo_ReadPositions(unsigned long channels, int cmd, unsigned long* width);
//-- values for the "cmd" argument of the above functions
     #define RCSERVO_READPOS_POWEROFF	        (0)
     #define RCSERVO_READPOS_USEPARAMETER1		(1)
     #define RCSERVO_READPOS_USEPARAMETER2		(2)
     #define RCSERVO_READPOS_USEPARAMETER3		(3)
     #define RCSERVO_READPOS_USEPARAMETER4		(4)
     #define RCSERVO_READPOS_USEPARAMETER5	    (5)
     #define RCSERVO_READPOS_USEPARAMETER6	    (6)
     #define RCSERVO_READPOS_USEPARAMETER7	    (7)

//-- some ad hoc functions only for use in multiprogramming OS
unsigned long rcservo_ReadPositionRB(int channel, int cmd);
unsigned long rcservo_ReadPositionDN(int channel, int cmd);
void          rcservo_ReadPositionsDN(unsigned long channels, int cmd, unsigned long* width);


//Functions for Play Mode:
void    rcservo_EnterPlayMode(void);

void rcservo_SetAction(unsigned long* width, unsigned long playtime);
void rcservo_PauseAction(void);
void rcservo_ReleaseAction(void);
int  rcservo_PlayAction(void);
//-- return values of rcservo_PlayAction()
     #define RCSERVO_PLAYEND (0)
     #define RCSERVO_PLAYING (1)
     #define RCSERVO_PAUSED  (2)

//-- parameter for controlling the behavior of rcservo_PlayAction()
extern unsigned long RCSERVO_playInterval;


//Functions for PWM Mode:
void rcservo_EnterPWMMode(void);
void rcservo_SendPWMPulses(int channel, unsigned long period, unsigned long duty, unsigned long count);
bool rcservo_IsPWMCompleted(int channel);


void rcservo_Outp(int channel, int value);  //no action if the "channel" is used for Servo Control
int rcservo_Inp(int channel);               //return 0xffff if the "channel" is used for Servo Control

#ifdef __cplusplus
}
#endif 

#endif
