// Roboard Demo.cpp : Defines the entry point for the console application.
//
#include <stdio.h>
#include <conio.h>
#include <stdlib.h>
#include <string.h>
#ifndef WIN32
    #define _stricmp  stricmp
    #define _strnicmp strnicmp
#endif

#include "common.h"
#include "demofile.h"
#include "io.h"
#include "rcservo.h"

#ifdef WIN32
	#include "fileio.h"
    #include <irrKlang.h>
    using namespace irrklang;

	class CMyFileFactory : public irrklang::IFileFactory
	{
	public:
		virtual irrklang::IFileReader* createFileReader(const ik_c8* filename)
		{
			FILEIO* file = fileio_FileOpen((char*)filename, "rb");
			if (!file) return 0;

			return new CMyReadFile(file, filename);
		}

	protected:
		class CMyReadFile : public irrklang::IFileReader
		{
		public:
			CMyReadFile(FILEIO* openedFile, const ik_c8* filename)
			{
				File = openedFile;
				strcpy(Filename, filename);
				FileSize = (ik_s32)fileio_GetSize(openedFile);
			}

			~CMyReadFile()
			{
				fileio_FileClose(File);
			}

			ik_s32 read(void* buffer, ik_u32 sizeToRead)
			{
				return (ik_s32)fileio_Read(buffer, 1, sizeToRead, File);
			}

			bool seek(ik_s32 finalPos, bool relativeMovement)
			{
				return fileio_Seek(File, finalPos, relativeMovement ? SEEK_CUR : SEEK_SET) == 0;
			}

			ik_s32 getSize()
			{
				return FileSize;
			}

			ik_s32 getPos()
			{
				return fileio_GetPos(File);
			}

			const ik_c8* getFileName()
			{
				return Filename;
			}

			FILEIO* File;
			char Filename[1024];
			ik_s32 FileSize;

		}; //end class CMyReadFile
	}; //end class CMyFileFactory
#endif


static bool init_rcservo(unsigned long channels) {
	int i;

	showmsg("Initializing RCSERVO lib...");
	for (i=0; i<32; i++) rcservo_SetServo(i, RCSERVO_KONDO_KRS78X);

	if (io_init() == true) //&&
	if (rcservo_Initialize(channels) == true)
	{
		showmsg("...success\n");
		RCSERVO_playInterval = 10L;
		return true;
	}

	showmsg("...fail!\n");
	errmsg("ERROR: RCSERVO lib fails to initialize (%s)!\n", ERR_Msg);
	return false;
}

static void close_rcservo(void) {
	showmsg("Closing RCSERVO lib...");

	rcservo_Close();
	io_close();

	showmsg("...done\n");
}

static unsigned long htol(char* str) {
    unsigned long value = 0L;

	for (int i=0; i<(int)strlen(str); i++)
		value = (str[i]>'9')? value*16+(unsigned long)(str[i]-'a'+10): value*16+(unsigned long)(str[i]-'0');

	return value;
}

static void writefile_frame(FILE* fp, unsigned long* width, int frameno) {
	int i;

	if (frameno != -1) fprintf(fp, "[Frame %d]\n", frameno);
	for (i=0; i<32; i++)
		fprintf(fp, "channel%d = %lu\n", i, width[i]);
	if (frameno != -1) fprintf(fp, "\n");
}


void mode_captureframes(char* filename, int frameno, unsigned long usedchannels) {
	unsigned long width[32];
	unsigned c = 0;
	int i;
	FILE* fp;

	showmsg("Open %s to write...", filename);
	if ((fp = fopen(filename, "r")) != NULL)
	{
		fclose(fp);
		showmsg("file already exists. overwirte? (Y/N): ");
		c = keyboard_waitkey();
		if ((c != 'y') && (c != 'Y')) return;
	}
	if ((fp = fopen(filename, "w")) == NULL)
	{
		showmsg("...fail!\n");
		return;
	}
	showmsg("...success\n");

	if (init_rcservo(usedchannels) == false) return; else showmsg("\n");
	for (i=0; i<32; i++) RCSERVO_servoParameters[i].feedbackReadMode = RCSERVO_READPOS_FASTMODE;

	showmsg("Pose the KONDO robot, and press any key to capture [ESC to quit] ...\n");
	rcservo_EnterCaptureMode();
	while ((c = keyboard_waitkey()) != KBESC)
	{
		rcservo_ReadPositionsDN(usedchannels, RCSERVO_READPOS_POWEROFF, width);
		for (i=0; i<32; i++)
		{
			if (width[i] == 0xffffffffL)
			{
				showmsg("ERROR: fail to read the position of channel %d!\n", i);
				width[i] = 0L;
			}
			else
				showmsg("The position of channel %d is %lu\n", i, width[i]);
		}//end for i...
		writefile_frame(fp, width, frameno);

		if (frameno == -1) break;
		showmsg("[Frame %d] captured\n\n", frameno); frameno++;
		showmsg("Pose the KONDO robot, and press any key to capture [ESC to quit] ...\n");
	}

	fclose(fp);
	close_rcservo();
}

void mode_replayframes(char* filename, int frameno, unsigned long replaytime, unsigned long usedchannels) {
	unsigned long* frmPool[999];
	int   i;
	bool  r;

	DEMOFILE_demoDirectory = "";
	demofile_InitFRM(frmPool, 999);
	if (frameno == -1)
	{
		showmsg("Read the frame from %s...", filename);
		r = demofile_ReadSingleFRM(filename, &(frmPool[0]));
	}
	else
	{
		showmsg("Read frames from %s...", filename);
		r = demofile_ReadFRM(filename, frmPool, 999);
	}

	if (r == false)
	{
		showmsg("...fail! (%s)\n", ERR_Msg);
		demofile_FreeFRM(frmPool, 999);
		return;
	}
	showmsg("...success\n");

	if (init_rcservo(usedchannels) == false) return; else showmsg("\n");

	rcservo_EnterPlayMode();
	for (i=(frameno==-1)?0:frameno; i<999; i++)
	{
		if (frmPool[i] == NULL) continue;

		showmsg("Playing [Frame %d]...", i);

		rcservo_SetAction(frmPool[i], replaytime);
		while (rcservo_PlayAction() != RCSERVO_PLAYEND);

		showmsg("...finish\n");
		if ((frameno == -1) || (i == 999-1))
		{
			showmsg("press any key to quit\n\n");
			keyboard_waitkey(); break;
		}
		else
		{
			showmsg("press any key to continue [ESC to quit]\n\n");
			if (keyboard_waitkey() == KBESC) break;
		}
	}

	showmsg("\nFinish the replay process.\n", filename);
	demofile_FreeFRM(frmPool, 999);

	close_rcservo();
}

void mode_playdemo(char* dirname, unsigned long idletime) {
	char  demodir[255];
	char  demodef[255];
	int   i;
	unsigned c;

	DEF_t  demo;
	ACT_t* curact;

	#define  NOPLAY		(0)
	#define  PLAYNEXT	(1)
	#define  PLAYING	(2)
	int  playing = NOPLAY;
	bool playpause;
	unsigned long idleTimeOut = timer_nowtime();
	unsigned long playframe[32];

	strcpy(demodir, dirname);
	strcat(demodir, "/");
	while (strcspn(dirname, "/\\") != strlen(dirname)) dirname = &dirname[strcspn(dirname, "/\\") + 1];
	strcpy(demodef, dirname);
	strcat(demodef, ".def");
	DEMOFILE_demoDirectory = demodir;

	showmsg("Loading demo: %s\n-----------------------------------\n", dirname);
	if (demofile_ReadDEF(demodef, &demo) == false)
	{
		showmsg("Fail to load (%s)!\n", ERR_Msg);
		return;
	}
	showmsg("-----------------------------------> completed\n");

	if (init_rcservo(demo.usedChannels) == false) return;
	for (i=0; i<32; i++)
	{
		RCSERVO_servoParameters[i].periodWidth  = demo.PWMPeriod;
		RCSERVO_servoParameters[i].minDutyWidth = demo.minPWMDuty;
		RCSERVO_servoParameters[i].maxDutyWidth = demo.maxPWMDuty;
	}

    #ifdef WIN32
	   showmsg("Initializing Audio...");
	   ISoundEngine* audioEngine = createIrrKlangDevice(ESOD_AUTO_DETECT, ESEO_MULTI_THREADED | ESEO_LOAD_PLUGINS);
	   if (!audioEngine)
		   showmsg("...fail to initialize audio!\n");
	   else
	   {
		   showmsg("...success\n\n");

		   //allow the audioEngine to read audio files in a zipped file
		   CMyFileFactory* demofileFactory = new CMyFileFactory();
		   audioEngine->addFileFactory(demofileFactory);
		   demofileFactory->drop();
	   }
	#endif

	showmsg("\nPress a command to play an action ([ESC] to quit)...\n");
	while ((c = keyboard_getkey()) != KBESC)
	{
		if (demo.initCMD != -1)
		{
			c = demo.cmdPool[demo.initCMD]->command;
			demo.initCMD = -1;
		}

		switch (playing)
		{
			case NOPLAY:
				if (timer_nowtime() >= idleTimeOut) rcservo_EnterCaptureMode();  //stop PWM output

				if (c == 0)
					break;
				else
				{
					for (i=0; i<demo.CMD_num; i++)
						if (c == demo.cmdPool[i]->command) break;
					if (i >= demo.CMD_num) break;
				}

				showmsg("Playing command: %s", demo.cmdPool[i]->name);
				curact = demo.actPool[demo.cmdPool[i]->actno];
				if (curact == NULL)
					showmsg("...WARNING: invalid action!\n");
				else
				{
					showmsg("\n");
					rcservo_EnterPlayMode();
					playing   = PLAYNEXT;
					playpause = false;
				}
				break;
			case PLAYNEXT:
				if (curact == NULL)
				{
					showmsg("Finish the action.\n");
					showmsg("\nPress a command to play an action ([ESC] to quit)...\n");
					playing = NOPLAY;
					idleTimeOut = timer_nowtime() + idletime;
					#ifdef WIN32
						if (audioEngine) audioEngine->removeAllSoundSources();
					#endif
					break;
				}

				switch (curact->type)
				{
					case ACT_PLAYFRAME:
						showmsg("  Move to an old-format frame...");
						for (i=0; i<32; i++) playframe[demo.channelMapping[i]] = curact->data.frame[i];
						rcservo_SetAction(playframe, curact->time);
						playing = PLAYING;
						break;
					case ACT_PLAYFRAMENO:
						showmsg("  Move to [Frame %d]...", curact->data.frameno);
						if (demo.frmPool[curact->data.frameno] == NULL)
							showmsg("WARNING: invalid frame!\n");
						else
						{
							for (i=0; i<32; i++) playframe[demo.channelMapping[i]] = demo.frmPool[curact->data.frameno][i];
							rcservo_SetAction(playframe, curact->time);
							playing = PLAYING;
						}
						break;
					case ACT_PAUSE:
						showmsg("  Pause %ld ms...", curact->time);
	                    for (i=0; i<32; i++) playframe[i] = 0L;
						rcservo_SetAction(playframe, curact->time);
						playing = PLAYING;
						break;
					case ACT_PLAYSND:
						showmsg("  Play Audio %s...", curact->data.filename);
						#ifdef WIN32
						  if (!audioEngine)
							 showmsg("no audio device!");
						  else
							 audioEngine->play2D(curact->data.filename, false, false, true);
						#else
                             showmsg("no audio support!");
						#endif
						showmsg("\n");
						break;
				}//end switch (curact->type)
				curact = curact->next;
				break;
			case PLAYING:
				if (c == KBEND)
				{
					showmsg("Stop\n");
					playing = PLAYNEXT;
					curact  = NULL;
					break;
				}
				else
				if (c == KBSPACE)
				{
					playpause = (playpause == false)? true : false;
					if (playpause == false)
					{
						showmsg("Play.");
						rcservo_ReleaseAction();
					}
					else
					{
						showmsg("Paused.");
						rcservo_PauseAction();
					}
				}

				if (rcservo_PlayAction() == RCSERVO_PLAYEND)
				{
					showmsg("\n");
					playing = PLAYNEXT;
				}
				break;
		}//end switch (playing)
	}//end while ((c = keyboard_getkey()...
	if (playing == PLAYING) showmsg("\n");

	demofile_FreeDEF(&demo);
	#ifdef WIN32
		if (audioEngine) audioEngine->drop();
	#endif
	close_rcservo();
}


int main(int argc, char* argv[])
{
	#ifdef WIN32
		DWORD Error, priClass;

		priClass = GetPriorityClass(GetCurrentProcess());
		if (!SetPriorityClass(GetCurrentProcess(), HIGH_PRIORITY_CLASS)) //REALTIME_PRIORITY_CLASS or HIGH_PRIORITY_CLASS
		{
			Error = GetLastError();
			errmsg("Error: failed to enter HIGH_PRIORITY mode (%d)!\n", Error);
			goto End;
		}
	#endif

	
	#define MODE_CAPTURE  (0)
	#define MODE_CAPTURES (2)
	#define MODE_REPLAY   (1)
	#define MODE_REPLAYS  (3)
	#define MODE_DEMO     (100)
	int mode = MODE_CAPTURE;

	char* filename = "0000.txt";
	int	  frameno  = 0;
	unsigned long replaytime   = 1000L; //1000ms
	unsigned long usedchannels = 0x00ffffe0L; //default to use channels 5~23


	//parse user's arguments
	int i, argc_all = argc;
	for (i=1, argc=1; i<argc_all; i++) if (argv[i][0] != '-') argc++; else break;

	if (argc > 1)
	{
		if ((strcmp(argv[1], "0") == 0) || (strcmp(argv[1], "CAPTURE") == 0))
		{
			mode = MODE_CAPTURE;
			if (argc > 2) filename = argv[2];
		}
		else
		if ((strcmp(argv[1], "00") == 0) || (strcmp(argv[1], "CAPTURES") == 0))
		{
			mode = MODE_CAPTURES;
			if (argc > 2) filename = argv[2];
			if (argc > 3) frameno = atoi(argv[3]);
		}
		else
		if ((strcmp(argv[1], "1") == 0) || (strcmp(argv[1], "REPLAY") == 0))
		{
			mode = MODE_REPLAY;
			if (argc > 2) filename = argv[2];
		}
		else
		if ((strcmp(argv[1], "11") == 0) || (strcmp(argv[1], "REPLAYS") == 0))
		{
			mode = MODE_REPLAYS;
			if (argc > 2) filename = argv[2];
			if (argc > 3) frameno = atoi(argv[3]);
		}
		else
		{
			mode = MODE_DEMO;
			filename   = argv[1];
			replaytime = 15000L;
		}
	}//end if (argc...

	for (i=argc; i<argc_all-1; i++)
	{
		if ((argv[i][0] != '-') || (argv[i+1][0] == '-')) continue;

		if ((_stricmp(argv[i], "-CH") == 0) || (_stricmp(argv[i], "-channels") == 0))
		{
			usedchannels = htol(argv[i+1]);
		}
		else
		if (_stricmp(argv[i], "-time") == 0)
		{
			replaytime = atol(argv[i+1]);
		}
	}//end for (i=argc...


	//setup log file
	if (err_SetLogFile("roboard.log") == false)
		showmsg("WARNING: can't open file roboard.log for logging error messages!\n");

	switch (mode)
	{
		case MODE_CAPTURE:
			mode_captureframes(filename, -1, usedchannels);
			break;
		case MODE_CAPTURES:
			mode_captureframes(filename, frameno, usedchannels);
			break;
		case MODE_REPLAY:
			mode_replayframes(filename, -1, replaytime, usedchannels);
			break;
		case MODE_REPLAYS:
			mode_replayframes(filename, frameno, replaytime, usedchannels);
			break;
		case MODE_DEMO:
			mode_playdemo(filename, replaytime);
			break;
	}

	err_CloseLogFile();
End:
	return 0;
}

